import { makeAutoObservable, runInAction } from "mobx";
import api, { CreddefRecordDto, SchemaRecordDto } from "@dashboard/engine-api";
import { toast } from "react-toastify";
import displayError from "@dashboard/utils/displayError";

class OfferCredentialDialogStore {
  public initLoading = false;
  public connectionId: string | undefined = undefined;
  public credDefList: CreddefRecordDto[] = [];
  public selectedCredDefId: string | null = null;
  public schemaLoading = false;
  public schema: SchemaRecordDto | null = null;

  public issuingLoading = false;

  constructor() {
    makeAutoObservable(this);
  }

  public async init(connectionId: string | undefined) {
    runInAction(() => {
      this.initLoading = true;
      this.connectionId = connectionId;
    });
    try {
      const credDefList = await api.fetchCredentialDefinitions();
      runInAction(() => {
        this.credDefList = credDefList;
        this.initLoading = false;
      });
    } catch (e: unknown) {
      displayError(e);
      runInAction(() => (this.initLoading = false));
    }
  }

  public selectCredDef = async (credentialDefId: string) => {
    runInAction(() => {
      this.selectedCredDefId = credentialDefId;
      this.schema = null;
      this.schemaLoading = true;
    });
    const credDef = this.credDefList.find((p) => p.id === credentialDefId);

    try {
      const schema = await api.getSchemaById({ id: credDef!.schemaId! });
      runInAction(() => {
        this.schema = schema;
        this.schemaLoading = false;
      });
    } catch (e: unknown) {
      displayError(e);
      runInAction(() => (this.schemaLoading = false));
    }
  };

  public offerCredential = async (
    attributes: { name: string; value: string }[],
  ) => {
    runInAction(() => (this.issuingLoading = true));
    try {
      const cred = await api.offerCredential({
        connectionId: this.connectionId,
        credentialDefinitionId: this.selectedCredDefId!,
        attributes: attributes,
      });
      toast.success("Credential offer sent");
      runInAction(() => {
        this.issuingLoading = false;
      });
      return cred;
    } catch (e: unknown) {
      displayError(e);
      runInAction(() => {
        this.issuingLoading = false;
      });
    }
    return null;
  };
}

export type { OfferCredentialDialogStore };

export default OfferCredentialDialogStore;
