import { makeAutoObservable, runInAction } from "mobx";
import tsaApi from "@dashboard/tsa-api";
import { toast } from "react-toastify";
import displayError from "@dashboard/utils/displayError";
import { ApiException } from "@dashboard/engine-api";

type SetToken = (token: string) => void;

class AuthFormStore {
  public loading = false;
  public acceptanceAwaiting = false;
  public acceptanceExpired = false;
  public mode: "email" | "vc" = "email";
  public proofRecordId: string | null = null;
  public proofUrlShort: string | null = null;
  public proofUrl: string | null = null;
  private _setToken: SetToken;

  constructor(setToken: SetToken) {
    this._setToken = setToken;
    makeAutoObservable(this);
  }

  public requestTsaAuthEmail = async (email: string) => {
    runInAction(() => (this.loading = true));
    try {
      const { message } = await tsaApi.loginLogin({ email });
      toast(message);

      runInAction(() => {
        this.loading = false;
      });
    } catch (e: unknown) {
      displayError(e);
      runInAction(() => (this.loading = false));
    }
  };

  public requestTsaProof = async () => {
    runInAction(() => {
      this.loading = true;
      this.acceptanceAwaiting = false;
      this.acceptanceExpired = false;
    });
    try {
      const { proofRecordId, proofUrl, proofUrlShort } =
        await tsaApi.loginLoginInvitation();
      runInAction(() => {
        this.loading = false;
        this.proofRecordId = proofRecordId;
        this.proofUrl = proofUrl;
        this.proofUrlShort = proofUrlShort;
      });
    } catch (e: unknown) {
      displayError(e);
      runInAction(() => (this.loading = false));
    }
    if (this.proofRecordId) {
      this._awaitProofAcceptance(this.proofRecordId);
    }
  };

  private _awaitProofAcceptance = async (proofRecordId: string) => {
    runInAction(() => (this.acceptanceAwaiting = true));
    try {
      const { token } = await tsaApi.loginLoginWithVC({
        proofRecordId,
      });
      this._setToken(token);
      runInAction(() => {
        this.acceptanceAwaiting = false;
      });
    } catch (e: unknown) {
      if (
        e instanceof ApiException ||
        (e as { isApiException: boolean }).isApiException
      ) {
        const apiExError = e as ApiException;
        if (apiExError.status !== 504) {
          displayError(e);
        }
      } else {
        displayError(e);
      }
      runInAction(() => {
        this.acceptanceAwaiting = false;
        this.acceptanceExpired = true;
      });
    }
  };
}

export type { AuthFormStore };

export default AuthFormStore;
