import {
  Agent,
  AutoAcceptCredential,
  AutoAcceptProof,
  BaseEvent,
  ConnectionsModule,
  CredentialEventTypes,
  CredentialsModule,
  CredentialState,
  CredentialStateChangedEvent,
  DidDocument,
  DidsModule,
  EncryptedMessage,
  JsonTransformer,
  Key,
  KeyDidResolver,
  KeyType,
  OutOfBandState,
  PeerDidResolver,
  ProofEventTypes,
  ProofExchangeRecord,
  ProofsModule,
  ProofState,
  ProofStateChangedEvent,
  TypedArrayEncoder,
  V2CredentialProtocol,
  V2ProofProtocol,
  W3cCredentialsModule,
  WalletError,
  WalletKeyExistsError,
  WebDidResolver,
  JwkDidResolver,
  DidRecord,
} from "@credo-ts/core";
import {
  AnonCredsCredentialFormatService,
  AnonCredsModule,
  AnonCredsProofFormatService,
} from "@credo-ts/anoncreds";
import {
  IndyVdrAnonCredsRegistry,
  IndyVdrIndyDidRegistrar,
  IndyVdrIndyDidResolver,
  IndyVdrModule,
} from "@credo-ts/indy-vdr";
import { indyVdr } from "@hyperledger/indy-vdr-nodejs";
import { anoncreds } from "@hyperledger/anoncreds-nodejs";
import { AskarModule } from "@credo-ts/askar";
import { ariesAskar } from "@hyperledger/aries-askar-nodejs";
import { Key as AskarKey, KeyAlgs } from "@hyperledger/aries-askar-shared";
import crypto from "crypto";
import {
  catchError,
  filter,
  lastValueFrom,
  map,
  Observable,
  BehaviorSubject,
  Subject,
  take,
  timeout,
} from "rxjs";
import {
  SvdxDidRegistrar,
  SvdxDidResolver,
  SvdxModule,
  DefaultIpfsClient,
  SvdxRecordsApi,
} from "@vereign/credo-did-svdx";

import { SubjectInboundTransport } from "./askar/transports/agent.subject.inbound.transport";
import { SubjectOutboundTransport } from "./askar/transports/agent.subject.outbound.transport";

export type EventBehaviorSubject = BehaviorSubject<BaseEvent>;
export type SubjectMessage = {
  message: EncryptedMessage;
  replySubject?: Subject<SubjectMessage>;
};
import { Request, Response, Express } from "express";
import url from "url";
import { JsonLdCredentialFormatService } from "./credo/JsonLdCredentialFormatService";
import { EntityNotFoundError } from "@ocm-engine/dtos";
import { createWalletKeyPairClass } from "@credo-ts/core/build/crypto/WalletKeyPair";

export const importDidsToWallet = async (
  agent: Agent,
  dids: Array<string>,
): Promise<void> => {
  for (const did in dids) {
    try {
      await agent.dids.import({
        did: dids[did],
        overwrite: false,
      });
    } catch (e) {
      console.log((e as Error).message);
    }
  }
};

export const generateKey = async ({
  seed,
  agent,
}: {
  seed: string;
  agent: Agent;
}): Promise<Key> => {
  if (!seed) {
    throw Error("No seed provided");
  }

  const seedBuffer = TypedArrayEncoder.fromString(seed);

  try {
    return await agent.wallet.createKey({
      seed: seedBuffer,
      keyType: KeyType.Ed25519,
    });
  } catch (e) {
    if (e instanceof WalletKeyExistsError) {
      const c = AskarKey.fromSeed({
        algorithm: KeyAlgs.Ed25519,
        seed: seedBuffer,
      });

      return Key.fromPublicKey(c.publicBytes, KeyType.Ed25519);
    }

    if (e instanceof WalletError) {
      throw new Error(`Failed to create key - ${e.message}`);
    }

    throw new Error("Unknown");
  }
};

export const generateDidWeb = async ({
  seed,
  agent,
  peerAddress,
}: {
  seed: string;
  agent: Agent;
  peerAddress: string;
}) => {
  console.log("Generating did web");
  const pubKey = await generateKey({ seed, agent });

  const parsedUrl = url.parse(peerAddress);
  let hostname = parsedUrl.hostname!;
  const port = parsedUrl.port;
  const pathname = parsedUrl.pathname?.replace(/^\/+|\/+$/g, "");

  // If port is specified, encode it
  if (port) {
    hostname += `%3A${port}`;
  }
  // Convert URLs to 'did:web' form
  let didWeb = `did:web:${hostname}`;
  if (pathname) {
    didWeb += `:${pathname.replace(/\//g, ":")}`;
  }

  const verificationMethodKey0Id = `${didWeb}#jwt-key0`;

  const jsonDidDoc = {
    "@context": [
      "https://www.w3.org/ns/did/v1",
      "https://w3id.org/security/suites/ed25519-2018/v1",
    ],
    id: didWeb,
    verificationMethod: [
      {
        id: verificationMethodKey0Id,
        type: "Ed25519VerificationKey2018",
        controller: didWeb,
        publicKeyBase58: pubKey.publicKeyBase58,
      },
    ],
    authentication: [verificationMethodKey0Id],
    assertionMethod: [verificationMethodKey0Id],
    keyAgreement: [verificationMethodKey0Id],
  };

  const didDocumentInstance = JsonTransformer.fromJSON(jsonDidDoc, DidDocument);

  const recordId = "did:web";
  const existingRecord = await agent.genericRecords.findById(recordId);
  if (existingRecord) {
    await agent.genericRecords.deleteById(recordId);
  }
  await agent.genericRecords.save({
    id: recordId,
    content: jsonDidDoc,
  });

  await agent.dids.import({
    did: didWeb,
    didDocument: didDocumentInstance,
    overwrite: false,
  });

  console.log("Generated did:web");
  console.log(didWeb);
  console.log(JSON.stringify(didDocumentInstance.toJSON(), null, 2));
};

export const generateDidSvdxForEmail = async (agent: Agent, email: string) => {
  const activeKeySeed = crypto.randomBytes(46).toString("hex");
  const nextKeySeed = crypto.randomBytes(46).toString("hex");
  const activeKey = await generateKey({ seed: activeKeySeed, agent });
  const nextKey = await generateKey({ seed: nextKeySeed, agent });

  const WalletKeyPair = createWalletKeyPairClass(agent.context.wallet);

  const keyPair = new WalletKeyPair({
    id: `key-1`,
    controller: `did:svdx:${activeKey.fingerprint}`,
    key: activeKey,
    wallet: agent.context.wallet,
  });

  const didResult = await agent.dids.create({
    method: "svdx",
    options: {
      inceptionKey: activeKey,
      nextKey: nextKey,
      signer: { signer: () => keyPair.signer() },
    },
  });

  if (!didResult.didState.didDocument || !didResult.didState.did) {
    throw new Error("Could not create did svdx");
  }

  const svdxApi: SvdxRecordsApi = agent.modules["svdx"];
  await svdxApi.save({
    email: email,
    did: didResult.didState.did,
    activeKey: activeKey.publicKeyBase58,
    activeKeySeed: activeKeySeed,
    nextKey: nextKey.publicKeyBase58,
    nextKeySeed: nextKeySeed,
  });

  return didResult;
};

export const generateDidSvdx = async ({
  seed,
  agent,
}: {
  seed: string;
  agent: Agent;
}) => {
  const activeKey = await generateKey({ seed, agent });
  const nextKey = await generateKey({ seed, agent }); // TODO Where to find next key?

  const WalletKeyPair = createWalletKeyPairClass(agent.context.wallet);

  const keyPair = new WalletKeyPair({
    id: `key-1`,
    controller: `did:svdx:${activeKey.fingerprint}`,
    key: activeKey,
    wallet: agent.context.wallet,
  });

  const didResult = await agent.dids.create({
    method: "svdx",
    options: {
      inceptionKey: activeKey,
      nextKey: nextKey,
      signer: { signer: () => keyPair.signer() },
    },
  });
  console.log(didResult);

  if (!didResult.didState.didDocument) {
    throw new Error("Could not create did svdx");
  }

  return didResult;
};

export const generateDidFromKey = (key: Key): string => {
  if (!key) {
    throw new Error("Key not provided");
  }

  return TypedArrayEncoder.toBase58(key.publicKey.slice(0, 16));
};

//eslint-disable-next-line
export const getAskarAnonCredsIndyModules = (
  networks: any,
  {
    ipfsUrl,
    ipfsBasicUser,
    ipfsBasicPass,
  }: {
    ipfsUrl: string;
    ipfsBasicUser: string;
    ipfsBasicPass: string;
  },
) => {
  const modules: {
    connections: ConnectionsModule;
    credentials: CredentialsModule<
      V2CredentialProtocol<
        (AnonCredsCredentialFormatService | JsonLdCredentialFormatService)[]
      >[]
    >;
    proofs: ProofsModule<V2ProofProtocol<AnonCredsProofFormatService[]>[]>;
    anoncreds: AnonCredsModule;
    indyVdr: IndyVdrModule;
    dids: DidsModule;
    askar: AskarModule;
    w3cCredentials: W3cCredentialsModule;
    svdx?: SvdxModule;
  } = {
    connections: new ConnectionsModule({
      autoAcceptConnections: true,
    }),
    credentials: new CredentialsModule({
      autoAcceptCredentials: AutoAcceptCredential.ContentApproved,
      credentialProtocols: [
        new V2CredentialProtocol({
          credentialFormats: [
            new AnonCredsCredentialFormatService(),
            new JsonLdCredentialFormatService(),
          ],
        }),
      ],
    }),
    proofs: new ProofsModule({
      autoAcceptProofs: AutoAcceptProof.ContentApproved,
      proofProtocols: [
        new V2ProofProtocol({
          proofFormats: [new AnonCredsProofFormatService()],
        }),
      ],
    }),
    anoncreds: new AnonCredsModule({
      registries: [new IndyVdrAnonCredsRegistry()],
      anoncreds,
      autoCreateLinkSecret: true,
    }),
    indyVdr: new IndyVdrModule({
      indyVdr,
      networks,
    }),
    dids: new DidsModule({
      registrars:
        ipfsUrl && ipfsBasicUser && ipfsBasicPass
          ? [new SvdxDidRegistrar(), new IndyVdrIndyDidRegistrar()]
          : [new IndyVdrIndyDidRegistrar()],
      resolvers: [
        new SvdxDidResolver(),
        new IndyVdrIndyDidResolver(),
        new KeyDidResolver(),
        new PeerDidResolver(),
        new WebDidResolver(),
        new JwkDidResolver(),
      ],
    }),
    askar: new AskarModule({
      ariesAskar,
    }),
    w3cCredentials: new W3cCredentialsModule(),
  };

  if (ipfsUrl && ipfsBasicUser && ipfsBasicPass) {
    modules.svdx = new SvdxModule({
      ipfsClient: new DefaultIpfsClient(ipfsUrl, ipfsBasicUser, ipfsBasicPass),
    });
  }

  return modules;
};

const INITIAL_EVENT: BaseEvent = {
  type: "unknown",
  payload: {},
  metadata: {
    contextCorrelationId: "-1",
  },
};

export const setupEventBehaviorSubjects = (
  agents: Agent[],
  eventTypes: string[],
): BehaviorSubject<BaseEvent>[] => {
  const behaviorSubjects: EventBehaviorSubject[] = [];

  for (const agent of agents) {
    const behaviorSubject = new BehaviorSubject<BaseEvent>(INITIAL_EVENT);

    for (const eventType of eventTypes) {
      agent.events.observable(eventType).subscribe(behaviorSubject);
    }

    behaviorSubjects.push(behaviorSubject);
  }

  return behaviorSubjects;
};

export const setupSubjectTransports = (agents: Agent[]) => {
  const subjectMap: Record<string, Subject<SubjectMessage>> = {};

  for (const agent of agents) {
    const messages = new Subject<SubjectMessage>();
    subjectMap[agent.config.endpoints[0]] = messages;
    agent.registerInboundTransport(new SubjectInboundTransport(messages));
    agent.registerOutboundTransport(new SubjectOutboundTransport(subjectMap));
  }
};

export const isProofStateChangedEvent = (
  e: BaseEvent,
): e is ProofStateChangedEvent => e.type === ProofEventTypes.ProofStateChanged;

export const isCredentialStateChangedEvent = (
  e: BaseEvent,
): e is CredentialStateChangedEvent =>
  e.type === CredentialEventTypes.CredentialStateChanged;

export const waitForProofExchangeRecordSubject = (
  subject: BehaviorSubject<BaseEvent> | Observable<BaseEvent>,
  {
    proofRecordId,
    threadId,
    parentThreadId,
    state,
    previousState,
    timeoutMs = 10000,
    count = 1,
  }: {
    proofRecordId?: string;
    threadId?: string;
    parentThreadId?: string;
    state?: ProofState;
    previousState?: ProofState | null;
    timeoutMs?: number;
    count?: number;
  },
): Promise<ProofExchangeRecord> => {
  const observable: Observable<BaseEvent> =
    subject instanceof BehaviorSubject ? subject.asObservable() : subject;
  return lastValueFrom(
    observable.pipe(
      filter(isProofStateChangedEvent),
      filter(
        (e) =>
          (proofRecordId === undefined ||
            e.payload.proofRecord.id === proofRecordId) &&
          (previousState === undefined ||
            e.payload.previousState === previousState) &&
          (threadId === undefined ||
            e.payload.proofRecord.threadId === threadId) &&
          (parentThreadId === undefined ||
            e.payload.proofRecord.parentThreadId === parentThreadId) &&
          (state === undefined || e.payload.proofRecord.state === state),
      ),
      timeout(timeoutMs),
      catchError(() => {
        throw new Error(
          `ProofStateChangedEvent event not emitted within specified timeout: ${timeoutMs}
          previousState: ${previousState},
          threadId: ${threadId},
          parentThreadId: ${parentThreadId},
          state: ${state}
        }`,
        );
      }),
      take(count),
      map((e) => e.payload.proofRecord),
    ),
  );
};

export const waitForCredentialExchangeRecordSubject = (
  subject: BehaviorSubject<BaseEvent> | Observable<BaseEvent>,
  {
    credentialRecordId,
    threadId,
    parentThreadId,
    state,
    previousState,
    timeoutMs = 10000,
    count = 1,
  }: {
    credentialRecordId?: string;
    threadId?: string;
    parentThreadId?: string;
    state?: CredentialState;
    previousState?: CredentialState | null;
    timeoutMs?: number;
    count?: number;
  },
) => {
  const observable: Observable<BaseEvent> =
    subject instanceof BehaviorSubject ? subject.asObservable() : subject;
  return lastValueFrom(
    observable.pipe(
      filter(isCredentialStateChangedEvent),
      filter(
        (e) =>
          (credentialRecordId === undefined ||
            e.payload.credentialRecord.id === credentialRecordId) &&
          (previousState === undefined ||
            e.payload.previousState === previousState) &&
          (threadId === undefined ||
            e.payload.credentialRecord.threadId === threadId) &&
          (parentThreadId === undefined ||
            e.payload.credentialRecord.parentThreadId === parentThreadId) &&
          (state === undefined || e.payload.credentialRecord.state === state),
      ),
      timeout(timeoutMs),
      catchError(() => {
        throw new Error(
          `CredentialStateChanged event not emitted within specified timeout: ${timeoutMs}
          previousState: ${previousState},
          threadId: ${threadId},
          parentThreadId: ${parentThreadId},
          state: ${state}
        }`,
        );
      }),
      take(count),
      map((e) => e.payload.credentialRecord),
    ),
  );
};

export const attachShortUrlHandler = (server: Express, agent: Agent): void => {
  server.get(
    "/invitations/:invitationId",
    async (req: Request, res: Response): Promise<void> => {
      try {
        const { invitationId } = req.params;

        const outOfBandRecord = await agent.oob.findByCreatedInvitationId(
          invitationId,
        );

        if (
          !outOfBandRecord ||
          outOfBandRecord.state !== OutOfBandState.AwaitResponse
        ) {
          res.status(404).send("Not found");
          return;
        }

        const invitationJson = outOfBandRecord.outOfBandInvitation.toJSON();

        res.header("Content-Type", "application/json").send(invitationJson);
        return;
      } catch (error) {
        console.error(error);
        res.status(500).send("Internal Server Error");
        return;
      }
    },
  );
};

export const attachDidWebHandler = (
  server: Express,
  agent: Agent,
  agentPeerAddress: string,
): void => {
  const parsedUrl = url.parse(agentPeerAddress);
  const pathname = parsedUrl.pathname?.replace(/^\/+|\/+$/g, "");

  let serverDidWebPath: string;
  if (pathname) {
    serverDidWebPath = `/did.json`;
  } else {
    serverDidWebPath = "/.well-known/did.json";
  }

  console.log("Listen did web requests on path " + serverDidWebPath);
  server.get(
    serverDidWebPath,
    async (req: Request, res: Response): Promise<void> => {
      try {
        const didWebRecord = await agent.genericRecords.findById("did:web");

        if (!didWebRecord) {
          res.status(404).send("Not found");
          return;
        }

        const didWebDoc = didWebRecord.content;

        res.header("Content-Type", "application/json").send(didWebDoc);
        return;
      } catch (error) {
        console.error(error);
        res.status(500).send("Internal Server Error");
        return;
      }
    },
  );
};

export const getFirstDidSvdxRecord = async (
  agent: Agent,
): Promise<DidRecord> => {
  const didWebs = await agent.dids.getCreatedDids({
    method: "svdx",
  });
  if (!didWebs.length) {
    throw new EntityNotFoundError("Agent does not have did:web");
  }
  return didWebs[0];
};
