import { Injectable, Logger } from "@nestjs/common";
import { AskarService } from "./askar.service";
import {
  CreateCredentialDefinitionRequestDto,
  CreddefRecordDto,
  CreateSchemaRequestDto,
  SchemaRecordDto,
  CredentialNotCreatedError,
  SchemaNotCreatedError,
  EntityNotFoundError,
} from "@ocm-engine/dtos";

@Injectable()
export class AgentAnoncredsService {
  private readonly logger = new Logger(AgentAnoncredsService.name);
  constructor(private readonly askar: AskarService) {}

  fetchSchemas = async (): Promise<SchemaRecordDto[]> => {
    let schemaRecords =
      await this.askar.agent.modules.anoncreds.getCreatedSchemas({});
    schemaRecords = schemaRecords.sort(
      (a, b) =>
        new Date(b.createdAt).getTime() - new Date(a.createdAt).getTime(),
    );

    const schemaResponses = schemaRecords.map((singleSchemaRes) => {
      return SchemaRecordDto.fromJson({
        id: singleSchemaRes.schemaId,
        createdAt: singleSchemaRes.createdAt,
        updatedAt: singleSchemaRes.updatedAt,
        name: singleSchemaRes.schema.name,
        attributes: singleSchemaRes.schema.attrNames,
        version: singleSchemaRes.schema.version,
        issuerId: singleSchemaRes.schema.issuerId,
        methodName: singleSchemaRes.methodName,
      });
    });

    return schemaResponses;
  };

  getSchemaById = async (schemaId: string): Promise<SchemaRecordDto> => {
    const agentResponse = await this.askar.agent.modules.anoncreds.getSchema(
      schemaId,
    );

    if (!agentResponse || !agentResponse.schema) {
      throw new EntityNotFoundError();
    }

    return SchemaRecordDto.fromJson({
      id: agentResponse.schemaId,
      name: agentResponse.schema.name,
      attributes: agentResponse.schema.attrNames,
      version: agentResponse.schema.version,
      issuerId: agentResponse.schema.issuerId,
    });
  };

  createSchema = async (
    schema: CreateSchemaRequestDto,
  ): Promise<SchemaRecordDto> => {
    const dids = await this.askar.agent.dids.getCreatedDids({ method: "indy" });

    const schemaResult =
      await this.askar.agent.modules.anoncreds.registerSchema({
        schema: {
          name: schema.name,
          issuerId: dids[0].did,
          attrNames: schema.attributes,
          version: schema.version,
        },
        options: {},
      });

    if (schemaResult.schemaState.state !== "finished") {
      throw new SchemaNotCreatedError();
    }

    return SchemaRecordDto.fromJson({
      id: schemaResult.schemaState.schemaId,
      name: schemaResult.schemaState.schema.name,
      attributes: schemaResult.schemaState.schema.attrNames,
      version: schemaResult.schemaState.schema.version,
      issuerId: schemaResult.schemaState.schema.issuerId,
    });
  };

  fetchCredentialDefinitions = async (): Promise<CreddefRecordDto[]> => {
    let credentialDefinitions =
      await this.askar.agent.modules.anoncreds.getCreatedCredentialDefinitions(
        {},
      );
    credentialDefinitions = credentialDefinitions.sort(
      (a, b) =>
        new Date(b.createdAt).getTime() - new Date(a.createdAt).getTime(),
    );

    const response: Array<CreddefRecordDto> = [];
    for (const credDef of credentialDefinitions) {
      const cd = CreddefRecordDto.fromJson({
        id: credDef.credentialDefinitionId,
        createdAt: credDef.createdAt,
        updatedAt: credDef.updatedAt,
        schemaId: credDef.credentialDefinition.schemaId,
        issuerId: credDef.credentialDefinition.issuerId,
        tag: credDef.credentialDefinition.tag,
      });
      response.push(cd);
    }

    return response;
  };

  getCredentialDefinitionById = async (
    credentialDefinitionId: string,
  ): Promise<CreddefRecordDto> => {
    const credDefs =
      await this.askar.agent.modules.anoncreds.getCreatedCredentialDefinitions({
        credentialDefinitionId,
      });

    const credDef = credDefs[0] || null;
    if (!credDef) {
      throw new EntityNotFoundError();
    }

    return CreddefRecordDto.fromJson({
      id: credDef.credentialDefinitionId,
      createdAt: credDef.createdAt,
      updatedAt: credDef.updatedAt,
      schemaId: credDef.credentialDefinition.schemaId,
      issuerId: credDef.credentialDefinition.issuerId,
      tag: credDef.credentialDefinition.tag,
    });
  };

  createCredentialDefinition = async (
    credentialDefinitionDto: CreateCredentialDefinitionRequestDto,
  ): Promise<CreddefRecordDto> => {
    const dids = await this.askar.agent.dids.getCreatedDids({ method: "indy" });

    const response =
      await this.askar.agent.modules.anoncreds.registerCredentialDefinition({
        credentialDefinition: {
          tag: credentialDefinitionDto.tag,
          issuerId: dids[0].did,
          schemaId: credentialDefinitionDto.schemaId,
        },
        options: {
          supportRevocation: credentialDefinitionDto.supportRevocation ?? false,
        },
      });

    if (response.credentialDefinitionState.state !== "finished") {
      throw new CredentialNotCreatedError();
    }

    return this.getCredentialDefinitionById(
      response.credentialDefinitionState.credentialDefinitionId,
    );
  };
}
