import { Injectable, Logger } from "@nestjs/common";
import { AskarService } from "./askar.service";
import {
  MakeBasicMessageRequestDto,
  MessageRecordDto,
  MessageFilterDto,
} from "@ocm-engine/dtos";
import {
  BasicMessageRecord,
  BasicMessageRole,
  ConnectionRecord,
  Query,
} from "@credo-ts/core";

@Injectable()
export class AgentBasicMessagesService {
  private readonly logger = new Logger(AgentBasicMessagesService.name);
  constructor(private readonly askar: AskarService) {}

  fetchBasicMessages = async (
    filter: MessageFilterDto,
  ): Promise<MessageRecordDto[]> => {
    const query: Query<BasicMessageRecord>[] = [];

    if (filter.role) {
      const roleQuery: Query<BasicMessageRecord> = {
        role: filter.role,
      };
      query.push(roleQuery);
    }

    if (filter.connectionId) {
      const connectionQuery: Query<BasicMessageRecord> = {
        connectionId: filter.connectionId,
      };
      query.push(connectionQuery);
    }

    let messages = await this.askar.agent.basicMessages.findAllByQuery({
      $and: query,
    });
    messages = messages.sort(
      (a, b) =>
        new Date(b.createdAt).getTime() - new Date(a.createdAt).getTime(),
    );

    const connectionIds = messages.map((message) => message.connectionId);

    const connections = await this.askar.agent.connections.findAllByQuery({
      $or: connectionIds.map((p) => ({ id: p })),
    });
    const grouppedConnections = connections.reduce(
      (acc: { [connId: string]: ConnectionRecord }, conn) => {
        acc[conn.id] = conn;
        return acc;
      },
      {},
    );

    const response: MessageRecordDto[] = [];
    for (const message of messages) {
      const connection = grouppedConnections[message.connectionId];
      const label = connection?.theirLabel || "";

      const dto = MessageRecordDto.fromJson({
        id: message.id,
        createdAt: message.createdAt,
        updatedAt: message.updatedAt,
        connectionId: message.connectionId,
        role: message.role,
        content: message.content,
        sentTime: message.sentTime,
        from: message.role === BasicMessageRole.Receiver ? label : "",
        to: message.role === BasicMessageRole.Sender ? label : "",
      });

      response.push(dto);
    }

    return response;
  };

  sendMessage = async (
    dto: MakeBasicMessageRequestDto,
  ): Promise<MessageRecordDto> => {
    const messageRecord = await this.askar.agent.basicMessages.sendMessage(
      dto.connectionId,
      dto.message,
    );

    const connRecord = await this.askar.agent.connections.findById(
      dto.connectionId,
    );

    return MessageRecordDto.fromJson({
      id: messageRecord.id,
      createdAt: messageRecord.createdAt,
      updatedAt: messageRecord.updatedAt,
      connectionId: messageRecord.connectionId,
      role: messageRecord.role,
      content: messageRecord.content,
      sentTime: messageRecord.sentTime,
      from: "",
      to: connRecord?.theirLabel || "",
    });
  };

  deleteMessageById = async (id: string): Promise<void> => {
    await this.askar.agent.basicMessages.deleteById(id);
  };
}
