import { Injectable, Logger } from "@nestjs/common";
import { AskarService } from "./askar.service";
import {
  BaseRecordDto,
  ConnectionRecordDto,
  EntityNotFoundError,
} from "@ocm-engine/dtos";
import { DidExchangeState } from "@credo-ts/core";

@Injectable()
export class AgentConnectionsService {
  private readonly logger = new Logger(AgentConnectionsService.name);
  constructor(private readonly askar: AskarService) {}

  async fetchConnections(): Promise<ConnectionRecordDto[]> {
    //TODO: no ordering in findAllByQuery
    const agentResponse = (await this.askar.agent.connections.getAll()).sort(
      (a, b) =>
        new Date(b.createdAt).getTime() - new Date(a.createdAt).getTime(),
    );

    const connectionArray = agentResponse.map((singleConnectionRes) => {
      return ConnectionRecordDto.fromJson({
        id: singleConnectionRes.id,
        createdAt: singleConnectionRes.createdAt,
        updatedAt: singleConnectionRes.updatedAt,
        state: singleConnectionRes.state,
        connectionName: singleConnectionRes.theirLabel,
        alias: singleConnectionRes.alias,
        did: singleConnectionRes.did,
        theirDid: singleConnectionRes.theirDid,
        invitationDid: singleConnectionRes.invitationDid,
        outOfBandId: singleConnectionRes.outOfBandId,
        imageUrl: singleConnectionRes.imageUrl,
      });
    });

    return connectionArray;
  }

  getConnectionByOobId = async (oobId: string) => {
    const connectionRecords =
      await this.askar.agent.connections.findAllByOutOfBandId(oobId);

    const connectionArray = connectionRecords.map((singleConnectionRes) => {
      return ConnectionRecordDto.fromJson({
        id: singleConnectionRes.id,
        createdAt: singleConnectionRes.createdAt,
        updatedAt: singleConnectionRes.updatedAt,
        state: singleConnectionRes.state,
        connectionName: singleConnectionRes.theirLabel,
        alias: singleConnectionRes.alias,
        did: singleConnectionRes.did,
        theirDid: singleConnectionRes.theirDid,
        invitationDid: singleConnectionRes.invitationDid,
        outOfBandId: singleConnectionRes.outOfBandId,
        imageUrl: singleConnectionRes.imageUrl,
      });
    });

    return connectionArray;
  };

  getConnectionById = async (id: string): Promise<ConnectionRecordDto> => {
    const agentResponse = await this.askar.agent.connections.findById(id);

    if (!agentResponse) {
      throw new EntityNotFoundError();
    }

    return ConnectionRecordDto.fromJson({
      id: agentResponse.id,
      state: agentResponse.state,
      connectionName: agentResponse.theirLabel,
      alias: agentResponse.alias,
      did: agentResponse.did,
      theirDid: agentResponse.theirDid,
      invitationDid: agentResponse.invitationDid,
      outOfBandId: agentResponse.outOfBandId,
      createdAt: agentResponse.createdAt,
      updatedAt: agentResponse.updatedAt,
      imageUrl: agentResponse.imageUrl,
    });
  };

  deleteConnectionById = async (id: string): Promise<void> => {
    const connection = await this.askar.agent.connections.getById(id);
    if (connection.theirDid) {
      await this.askar.agent.connections.hangup({
        connectionId: id,
      });
    }
    await this.askar.agent.connections.deleteById(id);
  };

  trustPingToConnection = async (connectionId: string) => {
    const msg = await this.askar.agent.connections.sendPing(connectionId, {
      responseRequested: true,
      withReturnRouting: false,
    });

    return BaseRecordDto.fromJson({
      id: msg.toJSON()["@id"],
    });
  };
}
