import { Injectable, Logger } from "@nestjs/common";
import { AskarService } from "./askar.service";
import {
  OfferCredentialRequestDto,
  CredentialRecordDto,
  CredentialFilterDto,
  CredentialOfferResponseDto,
  EntityNotFoundError,
  AcceptCredentialDto,
  CredentialFormatDataDto,
} from "@ocm-engine/dtos";
import {
  AutoAcceptCredential,
  CredentialExchangeRecord,
  CredentialState,
  Query,
} from "@credo-ts/core";
import { waitForCredentialExchangeRecordSubject } from "../../agent.utils";

@Injectable()
export class AgentCredentialsService {
  private readonly logger = new Logger(AgentCredentialsService.name);
  constructor(private readonly askar: AskarService) {}

  offerCredential = async (
    offerCredentialDto: OfferCredentialRequestDto,
  ): Promise<CredentialOfferResponseDto> => {
    this.logger.log(
      "Incoming request",
      JSON.stringify(offerCredentialDto, null, 2),
    );

    if (!offerCredentialDto.connectionId) {
      const { credentialRecord, message } =
        await this.askar.agent.credentials.createOffer({
          protocolVersion: "v2",
          credentialFormats: {
            anoncreds: {
              credentialDefinitionId: offerCredentialDto.credentialDefinitionId,
              attributes: offerCredentialDto.attributes,
            },
          },
          autoAcceptCredential: AutoAcceptCredential.ContentApproved,
        });

      credentialRecord.setTag("xRole", "issuer");
      await this.askar.agent.credentials.update(credentialRecord);

      const outOfBandRecord = await this.askar.agent.oob.createInvitation({
        messages: [message],
        handshake: false,
      });

      const credentialUrl = outOfBandRecord.outOfBandInvitation.toUrl({
        domain: this.askar.agentConfig.agentPeerAddress,
      });

      const shortCredentialUrl = `${this.askar.agentConfig.agentPeerAddress}/invitations/${outOfBandRecord.outOfBandInvitation.id}`;

      const dto = CredentialRecordDto.fromJson({
        id: credentialRecord.id,
        createdAt: credentialRecord.createdAt,
        updatedAt: credentialRecord.updatedAt,
        state: credentialRecord.state,
        connectionId: credentialRecord.connectionId,
        attributes: credentialRecord.credentialAttributes,
        tags: credentialRecord.getTags(),
      });

      return {
        credentialUrl: credentialUrl,
        shortCredentialUrl: shortCredentialUrl,
        credentialRecord: dto,
      };
    }
    const credentialExchangeRecord =
      await this.askar.agent.credentials.offerCredential({
        protocolVersion: "v2",
        connectionId: offerCredentialDto.connectionId,
        credentialFormats: {
          anoncreds: {
            credentialDefinitionId: offerCredentialDto.credentialDefinitionId,
            attributes: offerCredentialDto.attributes,
          },
        },
      });

    credentialExchangeRecord.setTag("xRole", "issuer");
    await this.askar.agent.credentials.update(credentialExchangeRecord);

    const dto = CredentialRecordDto.fromJson({
      id: credentialExchangeRecord.id,
      createdAt: credentialExchangeRecord.createdAt,
      updatedAt: credentialExchangeRecord.updatedAt,
      state: credentialExchangeRecord.state,
      connectionId: credentialExchangeRecord.connectionId,
      attributes: credentialExchangeRecord.credentialAttributes,
      tags: credentialExchangeRecord.getTags(),
    });

    return {
      credentialUrl: null,
      shortCredentialUrl: null,
      credentialRecord: dto,
    };
  };

  acceptCredential = async (
    acceptCredentialDto: AcceptCredentialDto,
  ): Promise<CredentialRecordDto> => {
    if (acceptCredentialDto.credentialUrl) {
      return this.acceptOobCredentials(acceptCredentialDto.credentialUrl);
    }
    return this.acceptConnectionCredential(acceptCredentialDto.credentialId);
  };

  acceptOobCredentials = async (url: string): Promise<CredentialRecordDto> => {
    // omit await in order to catch received record in the next line
    setTimeout(() => {
      this.askar.agent.oob.receiveInvitationFromUrl(url, {
        autoAcceptConnection: false,
        autoAcceptInvitation: true,
        // reuseConnection: true,
      });
    }, 20);

    const record = await waitForCredentialExchangeRecordSubject(
      this.askar.agentB,
      {
        state: CredentialState.OfferReceived,
      },
    );

    const acceptedRecord = await this.askar.agent.credentials.acceptOffer({
      credentialRecordId: record.id,
    });

    return CredentialRecordDto.fromJson({
      id: acceptedRecord.id,
      createdAt: acceptedRecord.createdAt,
      updatedAt: acceptedRecord.updatedAt,
      state: acceptedRecord.state,
      connectionId: acceptedRecord.connectionId,
      attributes: acceptedRecord.credentialAttributes,
      tags: acceptedRecord.getTags(),
    });
  };

  acceptConnectionCredential = async (
    credentialRecordId: string,
  ): Promise<CredentialRecordDto> => {
    const credentialExchangeRecord =
      await this.askar.agent.credentials.acceptOffer({
        credentialRecordId,
      });

    return CredentialRecordDto.fromJson({
      id: credentialExchangeRecord.id,
      createdAt: credentialExchangeRecord.createdAt,
      updatedAt: credentialExchangeRecord.updatedAt,
      state: credentialExchangeRecord.state,
      connectionId: credentialExchangeRecord.connectionId,
      attributes: credentialExchangeRecord.credentialAttributes,
      tags: credentialExchangeRecord.getTags(),
    });
  };

  declineCredential = async (
    credentialRecordId: string,
  ): Promise<CredentialRecordDto> => {
    const credentialExchangeRecord =
      await this.askar.agent.credentials.declineOffer(credentialRecordId);

    // send request to the issuer that the request is declined, to mark it as Abondoned
    await this.askar.agent.credentials.sendProblemReport({
      credentialRecordId: credentialRecordId,
      description: "Decline offer",
    });

    return CredentialRecordDto.fromJson({
      id: credentialExchangeRecord.id,
      createdAt: credentialExchangeRecord.createdAt,
      updatedAt: credentialExchangeRecord.updatedAt,
      state: credentialExchangeRecord.state,
      connectionId: credentialExchangeRecord.connectionId,
      attributes: credentialExchangeRecord.credentialAttributes,
      tags: credentialExchangeRecord.getTags(),
    });
  };

  fetchCredentials = async (
    filter: CredentialFilterDto,
  ): Promise<CredentialRecordDto[]> => {
    const query: Query<CredentialExchangeRecord>[] = [];

    if (filter.states) {
      const stateQuery: Query<CredentialExchangeRecord> = {
        $or: filter.states.map((state) => ({ state })),
      };
      query.push(stateQuery);
    }

    if (filter.connectionId) {
      const connectionQuery: Query<CredentialExchangeRecord> = {
        connectionId: filter.connectionId,
      };
      query.push(connectionQuery);
    }

    let credentials = await this.askar.agent.credentials.findAllByQuery({
      $and: query,
    });
    credentials = credentials.sort(
      (a, b) =>
        new Date(b.createdAt).getTime() - new Date(a.createdAt).getTime(),
    );

    const response: CredentialRecordDto[] = [];
    for (const offer of credentials) {
      const dto = CredentialRecordDto.fromJson({
        id: offer.id,
        createdAt: offer.createdAt,
        updatedAt: offer.updatedAt,
        state: offer.state,
        connectionId: offer.connectionId,
        attributes: offer.credentialAttributes,
        tags: offer.getTags(),
      });
      response.push(dto);
    }

    return response;
  };

  getCredentialById = async (
    credentialId: string,
  ): Promise<CredentialRecordDto> => {
    const credentialRecord = await this.askar.agent.credentials.findById(
      credentialId,
    );

    if (!credentialRecord) {
      throw new EntityNotFoundError();
    }

    return CredentialRecordDto.fromJson({
      id: credentialRecord.id,
      createdAt: credentialRecord.createdAt,
      updatedAt: credentialRecord.updatedAt,
      state: credentialRecord.state,
      connectionId: credentialRecord.connectionId,
      attributes: credentialRecord.credentialAttributes,
      tags: credentialRecord.getTags(),
    });
  };

  getCredentialFormatDataById = async (
    credentialId: string,
  ): Promise<CredentialFormatDataDto> => {
    const formatData = await this.askar.agent.credentials.getFormatData(
      credentialId,
    );

    if (!formatData) {
      throw new EntityNotFoundError();
    }

    return CredentialFormatDataDto.fromJson({
      proposalAttributes: formatData.proposalAttributes,
      offerAttributes: formatData.offerAttributes,
      anoncredsProposal: formatData.proposal?.anoncreds,
      anoncredsOffer: formatData.offer?.anoncreds,
      anoncredsRequest: formatData.request?.anoncreds,
      anoncredsCredential: formatData.credential?.anoncreds,
      all: formatData,
    });
  };

  deleteCredentialById = async (id: string): Promise<void> => {
    await this.askar.agent.credentials.deleteById(id);
  };
}
