import { Injectable, Logger } from "@nestjs/common";
import { AskarService } from "./askar.service";
import {
  CredentialRecordDto,
  CredentialOfferResponseDto,
  EntityNotFoundError,
  OcmError,
  W3cCredentialDto,
  W3cJsonLdVerifiableCredentialDto,
  W3cJsonLdVerifiablePresentationDto,
} from "@ocm-engine/dtos";
import {
  AutoAcceptCredential,
  ClaimFormat,
  JsonTransformer,
  W3cCredential,
  W3cCredentialService,
  JsonCredential,
  W3cJsonLdVerifiableCredential,
} from "@credo-ts/core";
import { uuid } from "@credo-ts/core/build/utils/uuid";
import { getFirstDidSvdxRecord } from "../../agent.utils";

@Injectable()
export class AgentJsonldService {
  private readonly logger = new Logger(AgentJsonldService.name);
  constructor(private readonly askar: AskarService) {}

  signJsonLdCredential = async (
    credToSign: W3cCredentialDto,
  ): Promise<W3cJsonLdVerifiableCredentialDto> => {
    this.logger.log("Sign json ld credentials");

    const didRecord = await getFirstDidSvdxRecord(this.askar.agent);
    const verificationMethodList =
      didRecord.didDocument?.verificationMethod || [];
    if (!verificationMethodList.length) {
      throw new EntityNotFoundError(
        "DidDocument does not exists or contains no verification methods",
      );
    }

    const verificationMethod = verificationMethodList[0];

    const w3cServ =
      this.askar.agent.context.dependencyManager.resolve(W3cCredentialService);

    credToSign.id = didRecord.did + "?uuid=" + uuid();
    if (credToSign.credentialSubject) {
      // eslint-disable-next-line @typescript-eslint/ban-ts-comment
      // @ts-ignore
      credToSign.credentialSubject.id = credToSign.id;
    }
    credToSign.issuer = didRecord.did;
    credToSign.issuanceDate = new Date().toISOString();
    let credential: W3cCredential;
    try {
      credential = JsonTransformer.fromJSON(credToSign, W3cCredential);
    } catch (e) {
      this.logger.log("Incorrect request parameter", e);
      throw new OcmError(
        "Invalid JSON-LD data format. Please ensure that your JSON-LD contains the following properties: @context, id, type, issuer, issuanceDate, expirationDate, and credentialSubject.",
      );
    }

    const vc = await w3cServ.signCredential(this.askar.agent.context, {
      format: ClaimFormat.LdpVc,
      credential,
      proofType: "Ed25519Signature2018",
      verificationMethod: verificationMethod.id,
    });

    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
    // @ts-ignore
    const jsonVC = vc.toJson() as W3cJsonLdVerifiableCredentialDto;
    this.logger.debug(JSON.stringify(jsonVC, null, 2));

    return jsonVC;
  };

  prepareVerifiablePresentationByJsonLdCredId = async (
    credentialRecordId: string,
  ): Promise<W3cJsonLdVerifiablePresentationDto> => {
    const didRecord = await getFirstDidSvdxRecord(this.askar.agent);
    const verificationMethodList =
      didRecord.didDocument?.verificationMethod || [];
    if (!verificationMethodList.length) {
      throw new EntityNotFoundError(
        "DidDocument does not exists or contains no verification methods",
      );
    }

    const verificationMethod = verificationMethodList[0];

    const credFormatData = await this.askar.agent.credentials.getFormatData(
      credentialRecordId,
    );
    if (!credFormatData.credential?.jsonld) {
      throw new OcmError(
        "The JSON-LD credential is either not in your wallet, pending approval, or not in the JSON-LD format.",
      );
    }

    const jsonLd = credFormatData.credential.jsonld;
    const vc = JsonTransformer.fromJSON(jsonLd, W3cJsonLdVerifiableCredential);

    const w3cServ =
      this.askar.agent.context.dependencyManager.resolve(W3cCredentialService);
    const presentation = await w3cServ.createPresentation({
      credentials: [vc],
      id: didRecord.did + "?uuid=" + uuid(),
    });

    const vp = await w3cServ.signPresentation(this.askar.agent.context, {
      format: ClaimFormat.LdpVp,
      presentation,
      proofPurpose: null,
      proofType: "Ed25519Signature2018",
      challenge: uuid(),
      verificationMethod: verificationMethod.id,
    });

    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
    // @ts-ignore
    const jsonVP = vp.toJson() as W3cJsonLdVerifiablePresentationDto;
    this.logger.debug(JSON.stringify(jsonVP, null, 2));

    return jsonVP;
  };

  offerJsonLdCredential = async (
    connectionId: string | undefined,
    credToSign: W3cCredentialDto,
  ): Promise<CredentialOfferResponseDto> => {
    this.logger.log("offerJsonLdCredential", connectionId);

    const didRecord = await getFirstDidSvdxRecord(this.askar.agent);
    const verificationMethodList =
      didRecord.didDocument?.verificationMethod || [];
    if (!verificationMethodList.length) {
      throw new EntityNotFoundError(
        "DidDocument does not exists or contains no verification methods",
      );
    }

    const verificationMethod = verificationMethodList[0];

    const w3cServ =
      this.askar.agent.context.dependencyManager.resolve(W3cCredentialService);

    credToSign.id = didRecord.did + "?uuid=" + uuid();
    if (credToSign.credentialSubject) {
      // eslint-disable-next-line @typescript-eslint/ban-ts-comment
      // @ts-ignore
      credToSign.credentialSubject.id = credToSign.id;
    }
    credToSign.issuer = didRecord.did;
    credToSign.issuanceDate = new Date().toISOString();
    let credential: W3cCredential;
    try {
      credential = JsonTransformer.fromJSON(credToSign, W3cCredential);
    } catch (e) {
      this.logger.log("Incorrect request parameter", e);
      throw new OcmError(
        "Invalid JSON-LD data format. Please ensure that your JSON-LD contains the following properties: @context, id, type, issuer, issuanceDate, expirationDate, and credentialSubject.",
      );
    }

    const vc = await w3cServ.signCredential(this.askar.agent.context, {
      format: ClaimFormat.LdpVc,
      credential,
      proofType: "Ed25519Signature2018",
      verificationMethod: verificationMethod.id,
    });
    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
    // @ts-ignore
    const jsonVC = vc.toJson() as JsonCredential;

    if (!connectionId) {
      // create connection less credential
      const { credentialRecord, message } =
        await this.askar.agent.credentials.createOffer({
          protocolVersion: "v2",
          credentialFormats: {
            jsonld: {
              credential: jsonVC,
              options: {
                proofType: "Ed25519Signature2018",
                proofPurpose: "assertionMethod",
              },
            },
          },
          autoAcceptCredential: AutoAcceptCredential.ContentApproved,
        });

      credentialRecord.setTag("xRole", "issuer");
      await this.askar.agent.credentials.update(credentialRecord);

      const outOfBandRecord = await this.askar.agent.oob.createInvitation({
        messages: [message],
        handshake: false,
      });

      const credentialUrl = outOfBandRecord.outOfBandInvitation.toUrl({
        domain: this.askar.agentConfig.agentPeerAddress,
      });

      const shortCredentialUrl = `${this.askar.agentConfig.agentPeerAddress}/invitations/${outOfBandRecord.outOfBandInvitation.id}`;

      const dto = CredentialRecordDto.fromJson({
        id: credentialRecord.id,
        createdAt: credentialRecord.createdAt,
        updatedAt: credentialRecord.updatedAt,
        state: credentialRecord.state,
        connectionId: credentialRecord.connectionId,
        attributes: credentialRecord.credentialAttributes,
        tags: credentialRecord.getTags(),
      });

      return {
        credentialUrl: credentialUrl,
        shortCredentialUrl: shortCredentialUrl,
        credentialRecord: dto,
      };
    }

    const credentialExchangeRecord =
      await this.askar.agent.credentials.offerCredential({
        connectionId: connectionId,
        protocolVersion: "v2",
        credentialFormats: {
          jsonld: {
            credential: jsonVC,
            options: {
              proofType: "Ed25519Signature2018",
              proofPurpose: "assertionMethod",
            },
          },
        },
        autoAcceptCredential: AutoAcceptCredential.ContentApproved,
      });

    this.logger.log(credentialExchangeRecord);

    credentialExchangeRecord.setTag("xRole", "issuer");
    await this.askar.agent.credentials.update(credentialExchangeRecord);

    const dto = CredentialRecordDto.fromJson({
      id: credentialExchangeRecord.id,
      createdAt: credentialExchangeRecord.createdAt,
      updatedAt: credentialExchangeRecord.updatedAt,
      state: credentialExchangeRecord.state,
      connectionId: credentialExchangeRecord.connectionId,
      attributes: credentialExchangeRecord.credentialAttributes,
      tags: credentialExchangeRecord.getTags(),
    });

    return {
      credentialUrl: null,
      shortCredentialUrl: null,
      credentialRecord: dto,
    };
  };
}
