import { Injectable, Logger } from "@nestjs/common";
import {
  EntityNotFoundError,
  RequestSenderEmailVcDto,
  RequestSenderEmailVcResponseDto,
} from "@ocm-engine/dtos";
import { CatalogClient } from "../clients/catalog.client";
import { generateDidSvdxForEmail } from "../../agent.utils";
import { AskarService } from "./askar.service";
import { IConfCatalog } from "@ocm-engine/config";
import { ConfigService } from "@nestjs/config";
import {
  ClaimFormat,
  JsonTransformer,
  W3cCredential,
  W3cCredentialService,
} from "@credo-ts/core";
import { uuid } from "@credo-ts/core/build/utils/uuid";

@Injectable()
export class AgentOcmService {
  private readonly logger = new Logger(AgentOcmService.name);
  private activated = false;
  constructor(
    private readonly askar: AskarService,
    private readonly catalogClient: CatalogClient,
    private readonly configService: ConfigService,
  ) {
    const catalogConfig = this.configService.get<IConfCatalog>("catalog")!;
    this.activated = !!catalogConfig.catalogUrl;
  }

  requestSenderEmailVC = async (
    dto: RequestSenderEmailVcDto,
  ): Promise<RequestSenderEmailVcResponseDto> => {
    this.ensureActivated();

    this.logger.debug(`requestSenderEmailVC`, dto.email);
    const didSvdx = await this.catalogClient.searchAccount(dto.email);
    if (didSvdx) {
      const record = await this.askar.agent.genericRecords.findById(
        "svdx_vc_" + dto.email,
      );
      if (!record) {
        // todo regenerate in that case?
        throw new Error(
          `VC not found in storage, even though the catalog has some data about it. did: ${didSvdx}, email: ${dto.email}`,
        );
      }
      const vc = record.content;
      return {
        email: dto.email,
        did: didSvdx,
        vc,
      };
    }

    const newDidSvdx = await generateDidSvdxForEmail(
      this.askar.agent,
      dto.email,
    );

    if (!newDidSvdx.didState.did) {
      throw new Error("Svdx did generation failed");
    }

    const didRecords = await this.askar.agent.dids.getCreatedDids({
      method: "svdx",
      did: newDidSvdx.didState.did,
    });
    if (!didRecords.length) {
      throw new EntityNotFoundError("Agent does not have did:svdx");
    }
    const svdxDid = didRecords[0];
    const verificationMethodList =
      svdxDid.didDocument?.verificationMethod || [];
    if (!verificationMethodList.length) {
      throw new EntityNotFoundError(
        "DidDocument does not exists or contains no verification methods",
      );
    }

    const verificationMethod = verificationMethodList[0];

    const credentialRecordId = uuid();
    const credential = JsonTransformer.fromJSON(
      {
        "@context": [
          "https://www.w3.org/2018/credentials/v1",
          "https://w3id.org/security/suites/jws-2020/v1",
          "https://www.vereign.com/.well-known/svdx-account-schema",
        ],
        type: ["VerifiableCredential"],
        id: svdxDid.did + "?uuid=" + credentialRecordId,
        issuer: svdxDid.did,
        issuanceDate: new Date().toISOString(),
        credentialSubject: {
          type: "svdx:SvdxAccount",
          "svdx:email": dto.email,
          "svdx:did": newDidSvdx.didState.did,
        },
      },
      W3cCredential,
    );

    const w3cServ =
      this.askar.agent.context.dependencyManager.resolve(W3cCredentialService);
    const vc = await w3cServ.signCredential(this.askar.agent.context, {
      format: ClaimFormat.LdpVc,
      credential,
      proofType: "Ed25519Signature2018",
      verificationMethod: verificationMethod.id,
    });
    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
    // @ts-ignore
    const jsonVC = vc.toJson();

    // Save credential
    await this.askar.agent.genericRecords.save({
      id: "svdx_vc_" + dto.email,
      content: jsonVC,
    });

    await this.catalogClient.createAccount(dto.email, svdxDid.did);

    return {
      email: dto.email,
      did: svdxDid.did,
      vc: jsonVC,
    };
  };

  private ensureActivated() {
    if (!this.activated) {
      throw new Error(
        "Please provide the catalog URL on startup to activate this endpoint.",
      );
    }
  }
}
