import { Injectable, Logger } from "@nestjs/common";
import { AskarService } from "./askar.service";
import {
  ConnectionRecordDto,
  CreateInvitationResponseDto,
  EntityNotFoundError,
  CreateInvitationRequestDto,
  InvitationFilterDto,
} from "@ocm-engine/dtos";
import { Query, OutOfBandRecord } from "@credo-ts/core";

@Injectable()
export class AgentOobService {
  private readonly logger = new Logger(AgentOobService.name);
  constructor(private readonly askar: AskarService) {}

  createInvitation = async (
    createInvitationRequestDto?: CreateInvitationRequestDto,
  ) => {
    const outOfBoundRecord = await this.askar.agent.oob.createInvitation(
      createInvitationRequestDto,
    );

    let longUrl = outOfBoundRecord.outOfBandInvitation.toUrl({
      domain: this.askar.agentConfig.agentPeerAddress,
    });

    if (this.askar.agentConfig.agentOobUrl) {
      longUrl = longUrl.replace(
        this.askar.agentConfig.agentPeerAddress,
        this.askar.agentConfig.agentOobUrl,
      );
    }

    //TODO: should we replace the short url with agentOobUrl if we do this we should have a redirect in ingress
    return CreateInvitationResponseDto.fromJson({
      createdAt: outOfBoundRecord.createdAt,
      updatedAt: outOfBoundRecord.updatedAt,
      outOfBandId: outOfBoundRecord.id,
      role: outOfBoundRecord.role,
      state: outOfBoundRecord.state,
      invitationUrl: longUrl,
      shortInvitationUrl: `${this.askar.agentConfig.agentPeerAddress}/invitations/${outOfBoundRecord.outOfBandInvitation.id}`,
    });
  };

  acceptInvitation = async (
    invitationUrl: string,
  ): Promise<ConnectionRecordDto> => {
    const { connectionRecord } =
      await this.askar.agent.oob.receiveInvitationFromUrl(invitationUrl);

    if (typeof connectionRecord === "undefined") {
      throw new EntityNotFoundError();
    }

    return ConnectionRecordDto.fromJson({
      connectionName: connectionRecord.theirLabel,
      state: connectionRecord.state,
      id: connectionRecord.id,
      did: connectionRecord.did,
      theirDid: connectionRecord.theirDid,
      invitationDid: connectionRecord.invitationDid,
      outOfBandId: connectionRecord.outOfBandId,
      createdAt: connectionRecord.createdAt,
      updatedAt: connectionRecord.updatedAt,
      imageUrl: connectionRecord.imageUrl,
    });
  };

  deleteInvitationById = async (id: string) => {
    return this.askar.agent.oob.deleteById(id);
  };

  fetchInvitations = async (filter: InvitationFilterDto) => {
    const query: Query<OutOfBandRecord>[] = [];

    if (filter.states) {
      const stateQuery: Query<OutOfBandRecord> = {
        $or: filter.states.map((state) => ({ state })),
      };
      query.push(stateQuery);
    }

    if (filter.roles) {
      const roleQuery: Query<OutOfBandRecord> = {
        $or: filter.roles.map((role) => ({ role })),
      };

      query.push(roleQuery);
    }

    let invitations = await this.askar.agent.oob.findAllByQuery({
      $and: query,
    });
    invitations = invitations.sort(
      (a, b) =>
        new Date(b.createdAt).getTime() - new Date(a.createdAt).getTime(),
    );

    const invitationsResponse = invitations.map((invitation) => {
      return CreateInvitationResponseDto.fromJson({
        createdAt: invitation.createdAt,
        updatedAt: invitation.updatedAt,
        outOfBandId: invitation.id,
        role: invitation.role,
        state: invitation.state,
        invitationUrl: invitation.outOfBandInvitation.toUrl({
          domain: this.askar.agentConfig.agentPeerAddress,
        }),
        shortInvitationUrl: `${this.askar.agentConfig.agentPeerAddress}/invitations/${invitation.outOfBandInvitation.id}`,
      });
    });

    return invitationsResponse;
  };

  getInvitationById = async (oobId: string) => {
    const invitation = await this.askar.agent.oob.getById(oobId);

    return CreateInvitationResponseDto.fromJson({
      createdAt: invitation.createdAt,
      updatedAt: invitation.updatedAt,
      outOfBandId: invitation.id,
      role: invitation.role,
      state: invitation.state,
      invitationUrl: invitation.outOfBandInvitation.toUrl({
        domain: this.askar.agentConfig.agentPeerAddress,
      }),
      shortInvitationUrl: `${this.askar.agentConfig.agentPeerAddress}/invitations/${invitation.outOfBandInvitation.id}`,
    });
  };
}
