import { Injectable, Logger } from "@nestjs/common";
import { AskarService } from "./askar.service";
import {
  AcceptProofDto,
  RequestProofDto,
  ProofRecordDto,
  ProofFilterDto,
  RequestProofResponseDto,
  EntityNotFoundError,
  ProofFormatDataDto,
} from "@ocm-engine/dtos";
import { ProofState, Query, ProofExchangeRecord } from "@credo-ts/core";
import { AnonCredsRequestedAttribute } from "@credo-ts/anoncreds";
import { uuid } from "@credo-ts/core/build/utils/uuid";
import { waitForProofExchangeRecordSubject } from "../../agent.utils";

@Injectable()
export class AgentProofsService {
  private readonly logger = new Logger(AgentProofsService.name);
  constructor(private readonly askar: AskarService) {}

  requestProof = async (
    requestProofDto: RequestProofDto,
  ): Promise<RequestProofResponseDto> => {
    this.logger.log(JSON.stringify(requestProofDto, null, 2));
    const requestedAttributes: Record<string, AnonCredsRequestedAttribute> = {};

    for (const attr of requestProofDto.attributes) {
      requestedAttributes[uuid()] = {
        name: attr.attributeName,
        restrictions: [
          {
            schema_id: attr.schemaId,
            cred_def_id: attr.credentialDefinitionId,
          },
        ],
      };
    }

    if (!requestProofDto.connectionId) {
      this.logger.log("connection Id not detected, creating oob proof");
      const { proofRecord, message } =
        await this.askar.agent.proofs.createRequest({
          protocolVersion: "v2",
          proofFormats: {
            anoncreds: {
              name: "proof-request",
              version: "1.0",
              requested_attributes: requestedAttributes,
            },
          },
        });

      proofRecord.setTag("xRole", "requester");
      await this.askar.agent.proofs.update(proofRecord);

      const outOfBandRecord = await this.askar.agent.oob.createInvitation({
        messages: [message],
        handshake: false,
      });

      const proofUrl = outOfBandRecord.outOfBandInvitation.toUrl({
        domain: this.askar.agentConfig.agentPeerAddress,
      });
      const shortProofUrl = `${this.askar.agentConfig.agentPeerAddress}/invitations/${outOfBandRecord.outOfBandInvitation.id}`;

      const dto = ProofRecordDto.fromJson({
        id: proofRecord.id,
        createdAt: proofRecord.createdAt,
        updatedAt: proofRecord.updatedAt,
        connectionId: proofRecord.connectionId,
        state: proofRecord.state,
        tags: proofRecord.getTags(),
      });

      return {
        proofUrl: proofUrl,
        shortProofUrl: shortProofUrl,
        proofRecord: dto,
      };
    }

    this.logger.log(`${requestProofDto.connectionId} detected, issuing proof`);

    const exchangeRecord = await this.askar.agent.proofs.requestProof({
      protocolVersion: "v2",
      connectionId: requestProofDto.connectionId,
      proofFormats: {
        anoncreds: {
          name: "proof-request",
          version: "1.0",
          requested_attributes: requestedAttributes,
        },
      },
    });

    exchangeRecord.setTag("xRole", "requester");
    await this.askar.agent.proofs.update(exchangeRecord);

    const dto = ProofRecordDto.fromJson({
      id: exchangeRecord.id,
      createdAt: exchangeRecord.createdAt,
      updatedAt: exchangeRecord.updatedAt,
      connectionId: exchangeRecord.connectionId,
      state: exchangeRecord.state,
      tags: exchangeRecord.getTags(),
    });

    return {
      proofUrl: null,
      shortProofUrl: null,
      proofRecord: dto,
    };
  };

  fetchProofs = async (filter: ProofFilterDto): Promise<ProofRecordDto[]> => {
    const query: Query<ProofExchangeRecord>[] = [];

    if (filter.states) {
      const stateQuery: Query<ProofExchangeRecord> = {
        $or: filter.states.map((state) => ({ state })),
      };
      query.push(stateQuery);
    }

    if (filter.connectionId) {
      const connectionQuery: Query<ProofExchangeRecord> = {
        connectionId: filter.connectionId,
      };
      query.push(connectionQuery);
    }

    let proofs = await this.askar.agent.proofs.findAllByQuery({
      $and: query,
    });
    proofs = proofs.sort(
      (a, b) =>
        new Date(b.createdAt).getTime() - new Date(a.createdAt).getTime(),
    );

    const response: ProofRecordDto[] = [];
    for (const proof of proofs) {
      const dto = ProofRecordDto.fromJson({
        id: proof.id,
        createdAt: proof.createdAt,
        updatedAt: proof.updatedAt,
        connectionId: proof.connectionId,
        state: proof.state,
        tags: proof.getTags(),
      });
      response.push(dto);
    }

    return response;
  };

  getProofById = async (proofRecordId: string): Promise<ProofRecordDto> => {
    const proofRecord = await this.askar.agent.proofs.findById(proofRecordId);

    if (!proofRecord) {
      throw new EntityNotFoundError();
    }

    return ProofRecordDto.fromJson({
      id: proofRecord.id,
      createdAt: proofRecord.createdAt,
      updatedAt: proofRecord.updatedAt,
      connectionId: proofRecord.connectionId,
      state: proofRecord.state,
      tags: proofRecord.getTags(),
    });
  };

  getProofFormatDataById = async (
    proofRecordId: string,
  ): Promise<ProofFormatDataDto> => {
    const formatData = await this.askar.agent.proofs.getFormatData(
      proofRecordId,
    );

    if (!formatData) {
      throw new EntityNotFoundError();
    }

    return ProofFormatDataDto.fromJson({
      anoncredsProposal: formatData.proposal?.anoncreds,
      anoncredsRequest: formatData.request?.anoncreds,
      anoncredsPresentation: formatData.presentation?.anoncreds,
    });
  };

  proofAcceptanceWait = async (
    proofRecordId: string,
  ): Promise<ProofFormatDataDto> => {
    const proofRecord = await this.askar.agent.proofs.findById(proofRecordId);

    if (!proofRecord) {
      throw new EntityNotFoundError();
    }

    if (proofRecord.state === ProofState.Done) {
      return this.getProofFormatDataById(proofRecordId);
    }

    await waitForProofExchangeRecordSubject(this.askar.agentB, {
      proofRecordId,
      state: ProofState.Done,
      timeoutMs: 3 * 60 * 1000, // 3 minutes
    });

    return this.getProofFormatDataById(proofRecordId);
  };

  acceptProof = async (
    acceptProofDto: AcceptProofDto,
  ): Promise<ProofRecordDto> => {
    if (acceptProofDto.proofUrl) {
      return this.acceptOobProof(acceptProofDto.proofUrl);
    }
    return this.acceptConnectionProof(acceptProofDto.proofId);
  };

  acceptOobProof = async (url: string): Promise<ProofRecordDto> => {
    // omit await in order to catch received record in the next line
    setTimeout(() => {
      this.askar.agent.oob.receiveInvitationFromUrl(url, {
        autoAcceptConnection: false,
        autoAcceptInvitation: true,
        // reuseConnection: true,
      });
    }, 20);

    const record = await waitForProofExchangeRecordSubject(this.askar.agentB, {
      state: ProofState.RequestReceived,
    });

    const requestedCredentials =
      await this.askar.agent.proofs.selectCredentialsForRequest({
        proofRecordId: record.id,
      });

    const acceptedRecord = await this.askar.agent.proofs.acceptRequest({
      proofRecordId: record.id,
      proofFormats: requestedCredentials.proofFormats,
    });

    return ProofRecordDto.fromJson({
      id: acceptedRecord.id,
      createdAt: acceptedRecord.createdAt,
      updatedAt: acceptedRecord.updatedAt,
      connectionId: acceptedRecord.connectionId,
      state: acceptedRecord.state,
      tags: acceptedRecord.getTags(),
    });
  };

  acceptConnectionProof = async (
    proofRecordId: string,
  ): Promise<ProofRecordDto> => {
    this.logger.log(`accepting proof request for ${proofRecordId}`);
    const requestedCredentials =
      await this.askar.agent.proofs.selectCredentialsForRequest({
        proofRecordId,
      });

    this.logger.log(JSON.stringify(requestedCredentials, null, 2));

    const proof = await this.askar.agent.proofs.acceptRequest({
      proofRecordId,
      proofFormats: requestedCredentials.proofFormats,
    });

    this.logger.log(JSON.stringify(proof, null, 2));

    return ProofRecordDto.fromJson({
      id: proof.id,
      createdAt: proof.createdAt,
      updatedAt: proof.updatedAt,
      connectionId: proof.connectionId,
      state: proof.state,
      tags: proof.getTags(),
    });
  };

  declineProofRequest = async (
    proofRecordId: string,
  ): Promise<ProofRecordDto> => {
    const resultFromDecline = await this.askar.agent.proofs.declineRequest({
      proofRecordId,
      sendProblemReport: true,
    });

    return ProofRecordDto.fromJson({
      id: resultFromDecline.id,
      createdAt: resultFromDecline.createdAt,
      updatedAt: resultFromDecline.updatedAt,
      connectionId: resultFromDecline.connectionId,
      state: resultFromDecline.state,
      tags: resultFromDecline.getTags(),
    });
  };

  deleteProofById = async (id: string): Promise<void> => {
    await this.askar.agent.proofs.deleteById(id);
  };
}
